' Copyright 2009 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

''' <summary>
''' This class stores and manages the tile providers
''' </summary>
Public Class TileProviders
    ''' <summary>
    ''' Datastructure of a tile provider
    ''' </summary>
    Public Structure TileProvider
        Public Name As String
        Public FolderName As String
        Public URL As String
        Public MinZoom As Integer
        Public MaxZoom As Integer
        Public Sub New(ByVal name As String, ByVal foldername As String, ByVal url As String, ByVal minZoom As Integer, ByVal maxZoom As Integer)
            Me.Name = name
            Me.FolderName = foldername
            Me.URL = url
            Me.MaxZoom = maxZoom
            Me.MinZoom = minZoom
        End Sub
    End Structure
    ''' <summary>
    ''' Array of tile providers
    ''' </summary>
    Private tileProviders(-1) As TileProvider
    ''' <summary>
    ''' Stores the index of the current/selected tile provider
    ''' </summary>
    Private iterator As Integer = -1

    ''' <summary>
    ''' Adds/inserts a new tile provider at the end.
    ''' </summary>
    ''' <param name="name">The displayed name of the tile provider.</param>
    ''' <param name="foldername">The foldername where to store/lookup the tiles.</param>
    ''' <param name="url">The URL where to downlaod the tiles.</param>
    ''' <param name="minZoom">The min zoom.</param>
    ''' <param name="maxZoom">The max zoom.</param>
    ''' <param name="interactive">if set to <c>true</c> if the process is happening interactive (i.e. the user can answer questions).</param>
    ''' <returns>
    ''' <c>true</c> if the insertion worked (i.e. the name is unique); otherwise, <c>false</c>.
    ''' </returns>
    Public Function AddTileProvider(ByVal name As String, ByVal foldername As String, ByVal url As String, ByVal minZoom As Integer, ByVal maxZoom As Integer, Optional ByVal interactive As Boolean = False) As Boolean
        If foldername.Length > 0 Then
            If Not IO.Directory.Exists(foldername) Then
                MsgBox("Tile folder missing, trying to create " & foldername, MsgBoxStyle.Information)
                Try
                    IO.Directory.CreateDirectory(foldername)
                Catch ex As Exception
                    MsgBox("Error creating folder: " & ex.Message & vbCrLf & "Disable tile storage for this provider.", MsgBoxStyle.Critical)
                    ' disable tile storage for this tileprovider
                    foldername = ""
                End Try
            End If
        End If
        If url.Length > 0 AndAlso url.StartsWith("http://") = False Then
            url = ""
        End If
        Dim isUniqueName As Integer = -1
        For i As Integer = 0 To tileProviders.Length - 1
            If tileProviders(i).Name = name Then
                isUniqueName = i
                Exit For
            End If
        Next
        If isUniqueName > -1 AndAlso interactive = False Then
            MsgBox("Tile provider """ + name + """ is not unique. Not added.", MsgBoxStyle.Critical)
            Return False
        ElseIf isUniqueName > -1 Then
            If MsgBox("Tile provider """ + name + """ is not unique. Overwrite?", MsgBoxStyle.Question Or MsgBoxStyle.YesNo) = MsgBoxResult.Yes Then
                tileProviders(isUniqueName).FolderName = foldername
                tileProviders(isUniqueName).URL = url
                tileProviders(isUniqueName).MinZoom = minZoom
                tileProviders(isUniqueName).MaxZoom = maxZoom
                Return True
            End If
        Return False
        End If
        ReDim Preserve tileProviders(tileProviders.Length)
        tileProviders(tileProviders.Length - 1) = New TileProvider(name, foldername, url, minZoom, maxZoom)
        If iterator = -1 Then
            iterator = 0
        End If
        Return True
    End Function
    ''' <summary>
    ''' Returns the path of the selected tile provider.
    ''' </summary>
    ''' <returns>The path where to store the tiles. Empty string if no provider exists.</returns>
    Public Function GetCurrentFolderName() As String
        If iterator = -1 Then
            Return ""
        Else
            Return tileProviders(iterator).FolderName
        End If
    End Function
    ''' <summary>
    ''' Returns the URL of the selected tile provider.
    ''' </summary>
    ''' <returns>The URL. Empty string if no provider exists.</returns>
    Public Function GetCurrentURL() As String
        If iterator = -1 Then
            Return ""
        Else
            Return tileProviders(iterator).URL
        End If
    End Function
    ''' <summary>
    ''' Gets the displayed name of the selected tile provider.
    ''' </summary>
    ''' <returns>The displayed name. Empty string if no provider exists.</returns>
    Public Function GetCurrentName() As String
        If iterator = -1 Then
            Return ""
        Else
            Return tileProviders(iterator).Name
        End If
    End Function
    ''' <summary>
    ''' Gets the min zoom provided by the selected tile provider.
    ''' </summary>
    ''' <returns>The min zoom provided by the tileprovider. "0" if no provider exists.</returns>
    Public Function GetCurrentMinZoom() As Integer
        If iterator = -1 Then
            Return 0
        Else
            Return tileProviders(iterator).MinZoom
        End If
    End Function
    ''' <summary>
    ''' Gets the max zoom provided by the selected tile provider.
    ''' </summary>
    ''' <returns>The min zoom provided by the tileprovider. "0" if no provider exists.</returns>
    Public Function GetCurrentMaxZoom() As Integer
        If iterator = -1 Then
            Return 0
        Else
            Return tileProviders(iterator).MaxZoom
        End If
    End Function
    ''' <summary>
    ''' Moves to the next tile provider in the list.
    ''' </summary>
    Public Sub NextProvider()
        If tileProviders.Length > 0 Then
            iterator += 1
            If iterator >= tileProviders.Length Then
                iterator = 0
            End If
        End If
    End Sub
    ''' <summary>
    ''' Removes all loaded tile providers.
    ''' </summary>
    Public Sub Clear()
        ReDim tileProviders(-1)
        iterator = -1
    End Sub
    ''' <summary>
    ''' Returns the selected tile provider.
    ''' </summary>
    ''' <returns>The tile provider. Empty tile provider if no provider exists.</returns>
    Public Function GetCurrentTileProvider() As TileProvider
        If iterator = -1 Then
            Return New TileProvider("(none)", "", "", 5, 5)
        Else
            Return tileProviders(iterator)
        End If
    End Function
    ''' <summary>
    ''' Returns the count of loaded tile providers.
    ''' </summary>
    ''' <returns>The count of loaded tile providers.</returns>
    Public Function Count() As Integer
        Return tileProviders.Length
    End Function
    ''' <summary>
    ''' Gotos to the provider at index <c>index</c>.
    ''' </summary>
    ''' <param name="index">The index.</param>
    ''' <returns>
    ''' <c>true</c> if the provider was selected; otherwise, <c>false</c>.
    ''' </returns>
    Public Function GotoProvider(ByVal index As Integer) As Boolean
        If tileProviders.Length > 0 Then
            If index >= 0 And index < tileProviders.Length Then
                iterator = index
                Return True
            End If
        End If
        Return False
    End Function
    ''' <summary>
    ''' Gotos to the provider with the displayed name <c>name</c>.
    ''' </summary>
    ''' <param name="name">The name.</param>
    ''' <returns>
    ''' <c>true</c> if the provider was found by its name and selected; otherwise, <c>false</c>.
    ''' </returns>
    Public Function GotoProvider(ByVal name As String) As Boolean
        For i As Integer = 0 To tileProviders.Length - 1
            If tileProviders(i).Name = name Then
                iterator = i
                Return True
            End If
        Next
        Return False
    End Function
    ''' <summary>
    ''' Deletes the selected tile provider.
    ''' </summary>
    Public Sub DeleteCurrent()
        If iterator >= 0 Then
            For i As Integer = iterator To tileProviders.Length - 2
                tileProviders(i) = tileProviders(i + 1)
            Next
            ReDim Preserve tileProviders(tileProviders.Length - 2)
            If tileProviders.Length = 0 Then
                iterator = -1
            Else
                iterator = 0
            End If
        End If
    End Sub
End Class
