' Copyright 2007 Rubke <interrubke@gmail.com>
' Copyright 2008 - 2010 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

Imports System.Net
Imports System.Runtime.InteropServices

''' <summary>
''' Control for displaying a map
''' </summary>
Public Class OSMTileMap

    Public Enum ScrollDirection As Integer
        UP
        DOWN
        RIGHT
        LEFT
    End Enum

    Public Event Zoomed()

    Private Const TILE_WIDTH As Integer = 256
    Private Const TILE_HEIGHT As Integer = 256

    Private tileCacheMutex As New Threading.Mutex
    Private smoothScrollingMutex As New Threading.Mutex

    Private currentTileProvider As TileProviders.TileProvider

    Private repaintMapHandle As IntPtr = IntPtr.Zero
    Private stopHandle As IntPtr = IntPtr.Zero
    Private repaintSlippyMapThread As Threading.Thread = Nothing
    Private myHalfWidth As Integer
    Private myHalfHeight As Integer
    Private resized As Boolean = False

    Private arrowHeight As Integer = 14
    Private arrowWidth As Integer = 8

    Private manualMode As Boolean = False
    Private mapPosition As New MAP_POSITION
    Private gpsPosition As New MAP_POSITION
    Private heading As Integer = 0

    Private nTileImage As Bitmap
    Private tiles(1, 1) As OSMTile
    Private blankTile As Bitmap = Nothing
    Private oldX As Integer = 0
    Private oldY As Integer = 0
    Private tileX As Integer
    Private tileY As Integer
    Private tileZ As Integer = 13
    Public GPXfile As GPXFile = Nothing
    Public OSMfile As OSMFile = Nothing
    Private drawTrackEnabled As Boolean = False
    Private paintPOIsEnabled As Boolean = False

    Private trackColor As Color = Color.Orange
    Private waypointColor As Color = Color.Blue

    'touchscreen map scrolling
    Private dragging As Boolean
    Private startX As Integer
    Private startY As Integer
    Private startPos As New PointD
    Private forceUpdate As Boolean = False ' ugly hack to force updating the Map even when no change in GPSpos
    Private paintWayPointsEnabled As Boolean = False
    Private downloadTilesEnabled As Boolean = False
    Private WithEvents downloader As OSMTileDownloader = New OSMTileDownloader

    ''' <summary>
    ''' Gets or sets the color of the gpx track.
    ''' </summary>
    ''' <value>The color of the gpx track.</value>
    Public Property PaintTrackColor() As Color
        Get
            Return trackColor
        End Get
        Set(ByVal value As Color)
            trackColor = value
        End Set
    End Property
    ''' <summary>
    ''' Gets or sets the color of the waypoints.
    ''' </summary>
    ''' <value>The color of the waypoints.</value>
    Public Property PaintWaypointColor() As Color
        Get
            Return waypointColor
        End Get
        Set(ByVal value As Color)
            waypointColor = value
        End Set
    End Property
    ''' <summary>
    ''' Gets or sets the size of the arrow which shows the current position and direction/heading.
    ''' </summary>
    ''' <value>The size (height in pixels) of the arrow.</value>
    Public Property ArrowSize() As Integer
        Get
            Return arrowHeight
        End Get
        Set(ByVal value As Integer)
            arrowHeight = value
            arrowWidth = CInt(value / 1.75)
        End Set
    End Property
    ''' <summary>
    ''' Gets or sets a value indicating whether the track (of GPXFile) should be painted on the map.
    ''' </summary>
    ''' <value><c>true</c> if the track should be painted on the map; otherwise, <c>false</c>.</value>
    Public Property DrawTrack() As Boolean
        Get
            Return drawTrackEnabled
        End Get
        Set(ByVal value As Boolean)
            drawTrackEnabled = value
        End Set
    End Property
    ''' <summary>
    ''' Gets or sets a value indicating whether POIs should be painted on the map.
    ''' </summary>
    ''' <value><c>true</c> if POIs should be painted; otherwise, <c>false</c>.</value>
    Public Property PaintPOIs() As Boolean
        Get
            Return paintPOIsEnabled
        End Get
        Set(ByVal value As Boolean)
            paintPOIsEnabled = value
        End Set
    End Property
    ''' <summary>
    ''' Gets or sets a value indicating whether downloading tiles is enabled or not.
    ''' </summary>
    ''' <value><c>true</c> if tiles download is enabled; otherwise, <c>false</c>.</value>
    Public Property DownloadTiles() As Boolean
        Get
            DownloadTiles = downloadTilesEnabled
        End Get
        Set(ByVal value As Boolean)
            downloadTilesEnabled = value
            If downloadTilesEnabled = False Then
                downloader.StopDownloader()
            End If
        End Set
    End Property
    ''' <summary>
    ''' Gets or sets the current zoomlevel of the map.
    ''' </summary>
    ''' <value>The current zoom.</value>
    Public Property CurrentZoom() As Integer
        Get
            CurrentZoom = tileZ
        End Get
        Set(ByVal value As Integer)
            If value >= currentTileProvider.MinZoom AndAlso value <= currentTileProvider.MaxZoom Then
                tileZ = value
                'UpdateMap()
            End If
        End Set
    End Property
    ''' <summary>
    ''' Sets the map position to display (also enters manual scrolling mode).
    ''' </summary>
    ''' <param name="newPosition">The new position to show on the map.</param>
    ''' <param name="visible">if set to <c>true</c> the map is visible and must be refreshed/recalculated.</param>
    Public Sub SetMapPosition(ByVal newPosition As MAP_POSITION, Optional ByVal visible As Boolean = True)
        EnterManualMode()

        If visible = True Then
            If newPosition <> mapPosition Then
                mapPosition = newPosition
                UpdateMap()
            End If
        End If
    End Sub
    ''' <summary>
    ''' Sets the GPS position to dispaly the current position on the map and updates the map if the map is visible.
    ''' </summary>
    ''' <param name="newGPSPosition">The new GPS position.</param>
    ''' <param name="visible">if set to <c>true</c> the map is visible and must be refreshed/recalculated.</param>
    Public Sub SetGPSPosition(ByRef newGPSPosition As GPS_POSITION, Optional ByVal visible As Boolean = True)
        If visible = True Then
            If newGPSPosition.Speed >= 0.8 Then
                heading = CInt(newGPSPosition.Heading)
            End If
            Dim newPosition As New MAP_POSITION(newGPSPosition)
            If newPosition <> gpsPosition Then
                gpsPosition = newPosition
                UpdateMap()
            Else
                Me.Refresh()
            End If
        End If
    End Sub
    Protected Overrides Sub OnPaint(ByVal pe As System.Windows.Forms.PaintEventArgs)
        Dim rc As Rectangle = Me.ClientRectangle

        If (nTileImage Is Nothing) Then ' Bitmap for doublebuffering
            pe.Graphics.FillRectangle(New SolidBrush(Color.Black), 0, 0, Me.Width - 1, Me.Height - 1)
        Else
            ' Check non-blocking if nTileImage is updated right now
            If smoothScrollingMutex.WaitOne(0, False) = False Then
                ' Consider painting it black or print some information here
                Exit Sub
            End If
            pe.Graphics.DrawImage(nTileImage, myHalfWidth - oldX, myHalfHeight - oldY)

            If manualMode = False Then
                DrawPosition(pe.Graphics, myHalfWidth, myHalfHeight)
            Else
                ' this costs much calculation power, not good to do it here, because OnPaint might get called very often!
                Dim x_pos As Integer, y_pos As Integer
                GetXYPosition(gpsPosition, tileX, tileY, x_pos, y_pos)
                x_pos = myHalfWidth + (x_pos - oldX)
                y_pos = myHalfHeight + (y_pos - oldY)
                If x_pos > 0 AndAlso y_pos > 0 AndAlso x_pos < Me.Width AndAlso y_pos < Me.Height Then
                    DrawPosition(pe.Graphics, x_pos, y_pos)
                End If
            End If

            Dim fnt As Font = New Font("arial", 7, FontStyle.Regular)
            Dim str As String = currentTileProvider.Name & ", Zoom: " & tileZ
            Dim s As SizeF = pe.Graphics.MeasureString(str, fnt)
            pe.Graphics.DrawString(str, fnt, New SolidBrush(Color.Green), Me.Width - s.Width - 2, Me.Height - s.Height - 2)
            smoothScrollingMutex.ReleaseMutex()
        End If
    End Sub
    Protected Overrides Sub OnPaintBackground(ByVal pe As System.Windows.Forms.PaintEventArgs)
        ' needs to be empty, otherwise you see the map flashing
    End Sub
    ''' <summary>
    ''' A resize handler which informs the PrepareSlippyMapThread to update the nTileImage and the map
    ''' </summary>
    ''' <param name="sender">ignored</param>
    ''' <param name="e">ignored</param>
    Private Sub ResizeHandler(ByVal sender As Object, ByVal e As System.EventArgs) Handles MyBase.Resize
        myHalfWidth = Me.Width >> 1
        myHalfHeight = Me.Height >> 1
        resized = True
        UpdateMap()
    End Sub
    ''' <summary>
    ''' Background thread which prepares the slippy map and draws it to nTileImage.
    ''' </summary>
    Private Sub PrepareSlippyMapThread()
        Dim abort As Boolean = False
        Dim multiHandler As IntPtr = LocalAlloc(LMEM_ZEROINIT, IntPtr.Size * 2)
        Marshal.WriteInt32(multiHandler, 0, stopHandle.ToInt32())
        Marshal.WriteInt32(multiHandler, IntPtr.Size, repaintMapHandle.ToInt32())
        Dim g As Graphics = Nothing
        Dim mapCenterPos As MAP_POSITION
        While Not abort

            If manualMode = True Then
                ' if in manual mode use our manual map position
                mapCenterPos = mapPosition
            Else
                ' otherwise in automaic mode track the gps position
                mapCenterPos = gpsPosition
            End If
            ResetEvent(repaintMapHandle)
            If resized Then
                'create 1 big slippy map canvas
                resized = False
                Dim x As Integer = 0
                Dim y As Integer = 0
                x = CInt(Fix(1 + (myHalfWidth << 1) / TILE_WIDTH))
                y = CInt(Fix(1 + (myHalfHeight << 1) / TILE_HEIGHT))
                If y <> tiles.GetLength(0) OrElse x <> tiles.GetLength(1) OrElse nTileImage Is Nothing Then
                    tileCacheMutex.WaitOne()
                    For i As Integer = 0 To tiles.GetLength(0) - 1
                        For j As Integer = 0 To tiles.GetLength(1) - 1
                            If tiles(i, j) IsNot Nothing Then
                                tiles(i, j).Dispose()
                            End If
                        Next j
                    Next i
                    ReDim tiles(y, x)
                    For i As Integer = 0 To tiles.GetLength(0) - 1
                        For j As Integer = 0 To tiles.GetLength(1) - 1
                            tiles(i, j) = New OSMTile()
                        Next j
                    Next i
                    smoothScrollingMutex.WaitOne()
                    tileCacheMutex.ReleaseMutex()
                    If nTileImage IsNot Nothing Then
                        nTileImage.Dispose()
                    End If
                    nTileImage = New Bitmap((x + 1) * TILE_WIDTH, (y + 1) * TILE_HEIGHT, Imaging.PixelFormat.Format16bppRgb565)
                    If g IsNot Nothing Then
                        g.Dispose()
                    End If
                    g = Graphics.FromImage(nTileImage)
                    smoothScrollingMutex.ReleaseMutex()
                    forceUpdate = True
                End If
            End If

            Dim xPos As Integer, yPos As Integer
            GetXYPosition(mapCenterPos, tileX, tileY, xPos, yPos)
            smoothScrollingMutex.WaitOne()
            Try
                forceUpdate = forceUpdate = True OrElse _
                   Not (nTileImage.Width - xPos > myHalfWidth AndAlso nTileImage.Height - yPos > myHalfHeight AndAlso _
                   myHalfWidth - xPos <= 0 AndAlso myHalfHeight - yPos <= 0)
            Finally
                smoothScrollingMutex.ReleaseMutex()
            End Try
            If Not forceUpdate Then
                oldX = xPos
                oldY = yPos
                ' draw new point
                If drawTrackEnabled = True Then
                    If manualMode = True Then
                        ' if manual mode is on, trackpoint are not on the center of map
                        ' so we need a new calculation based on gps position
                        GetXYPosition(gpsPosition, tileX, tileY, xPos, yPos)
                    End If
                    ' only if we are going to draw on preloaded "tile-canvas"
                    If xPos > 0 AndAlso yPos > 0 AndAlso xPos < nTileImage.Width AndAlso yPos < nTileImage.Height Then
                        If paintWayPointsEnabled Then
                            g.FillEllipse(New SolidBrush(waypointColor), xPos, yPos, 3, 3)
                            paintWayPointsEnabled = False
                        Else
                            g.FillEllipse(New SolidBrush(trackColor), xPos, yPos, 3, 3)
                        End If
                    End If
                End If
            Else
                forceUpdate = False

                Dim tileXDiff As Integer = tileX
                Dim tileYDiff As Integer = tileY

                CalcTileXY(mapCenterPos, tileZ, tileX, tileY)
                GetXYPosition(mapCenterPos, tileX, tileY, xPos, yPos)

                xPos = myHalfWidth - xPos
                yPos = myHalfHeight - yPos

                tileX -= CInt(Fix(1 + xPos / TILE_WIDTH))
                tileY -= CInt(Fix(1 + yPos / TILE_HEIGHT))

                ' prevent OnPaint to show wrong images and/or wrong positions
                smoothScrollingMutex.WaitOne()

                Try
                    GetXYPosition(mapCenterPos, tileX, tileY, oldX, oldY)

                    tileCacheMutex.WaitOne()
                    Try

                        tileXDiff = tileX - tileXDiff
                        tileYDiff = tileY - tileYDiff

                        If ((tileXDiff <> 0 OrElse tileYDiff <> 0) AndAlso (tileXDiff > -tiles.GetLength(1) + 1 AndAlso tileXDiff < tiles.GetLength(1) - 1) AndAlso (tileYDiff > -tiles.GetLength(0) + 1 AndAlso tileYDiff < tiles.GetLength(0) - 1)) Then
                            Dim newTiles(tiles.GetLength(0) - 1, tiles.GetLength(1) - 1) As OSMTile
                            For i As Integer = 0 To tiles.GetLength(0) - 1
                                For j As Integer = 0 To tiles.GetLength(1) - 1
                                    newTiles(PositiveMod(i - tileYDiff, tiles.GetLength(0)), PositiveMod(j - tileXDiff, tiles.GetLength(1))) = tiles(i, j)
                                Next
                            Next
                            tiles = newTiles
                        End If

                        For i As Integer = 0 To tiles.GetLength(0) - 1
                            For j As Integer = 0 To tiles.GetLength(1) - 1
                                If Not tiles(i, j).CompareCoordinates(tileX + j, tileY + i, tileZ) Then
                                    tiles(i, j).SetTile(tileX + j, tileY + i, tileZ, RetrieveTile(PositiveMod(tileX + j, PowerOf2(tileZ)), PositiveMod(tileY + i, PowerOf2(tileZ)), tileZ, downloader))
                                End If
                                If tiles(i, j).IsAvailable Then
#If Win32 Then
                                    g.DrawImage(tiles(i, j).GetTileImage, New Rectangle(j * (TILE_WIDTH), i * (TILE_HEIGHT), 256, 256), New Rectangle(0, 0, 256, 256), GraphicsUnit.Pixel)
#Else
                                    g.DrawImage(tiles(i, j).GetTileImage, j * (TILE_WIDTH), i * (TILE_HEIGHT))
#End If
                                Else
                                    g.DrawImage(blankTile, j * TILE_WIDTH, i * TILE_HEIGHT)
                                End If
                            Next
                        Next
                    Finally
                        tileCacheMutex.ReleaseMutex()
                    End Try
                Finally
                    smoothScrollingMutex.ReleaseMutex()
                End Try
                If paintPOIsEnabled AndAlso OSMfile IsNot Nothing Then
                    ' osm layer
                    OSMfile.drawOsmLayer(g, tileX, tileY, tileZ, nTileImage.Width, nTileImage.Height)
                End If

                If drawTrackEnabled = True AndAlso GPXfile IsNot Nothing Then
                    Dim x_pos As Integer, y_pos As Integer
                    Dim oldi_x As Integer, oldi_y As Integer
                    Dim trackpointlist As SingleLinkedList(Of GPXFile.OSMTrackPoint) = GPXfile.GetTrackPoints
                    If trackpointlist IsNot Nothing Then
                        Static old_z As Integer = 0
                        Dim paintTrackPen As New Pen(trackColor, 3)
                        Dim enumerator As IEnumerator(Of GPXFile.OSMTrackPoint) = GPXfile.GetTrackPoints.GetEnumerator
                        If enumerator.MoveNext() Then
                            If enumerator.Current.tilePosX = 0 OrElse old_z <> tileZ Then
                                GetXYPosition(enumerator.Current.Position, tileX, tileY, oldi_x, oldi_y)
                                enumerator.Current.SetCache(oldi_x + tileX * TILE_WIDTH, oldi_y + tileY * TILE_HEIGHT, False)
                            Else
                                oldi_x = enumerator.Current.tilePosX - tileX * TILE_WIDTH
                                oldi_y = enumerator.Current.tilePosY - tileY * TILE_HEIGHT
                            End If
                        End If
                        While enumerator.MoveNext
                            If enumerator.Current.tilePosX = 0 OrElse old_z <> tileZ Then
                                GetXYPosition(enumerator.Current.Position, tileX, tileY, x_pos, y_pos)
                                If Math.Abs(oldi_x - x_pos) > 3 OrElse Math.Abs(oldi_y - y_pos) > 3 Then
                                    enumerator.Current.SetCache(x_pos + tileX * TILE_WIDTH, y_pos + tileY * TILE_HEIGHT, False)
                                Else
                                    enumerator.Current.SetCache(x_pos + tileX * TILE_WIDTH, y_pos + tileY * TILE_HEIGHT, True)
                                End If
                            Else
                                x_pos = enumerator.Current.tilePosX - tileX * TILE_WIDTH
                                y_pos = enumerator.Current.tilePosY - tileY * TILE_HEIGHT
                            End If
                            If Not enumerator.Current.skipPaint Then
                                g.DrawLine(paintTrackPen, oldi_x, oldi_y, x_pos, y_pos)
                                oldi_x = x_pos
                                oldi_y = y_pos
                            End If
                        End While
                        old_z = tileZ
                    End If

                    Dim wayPointList As SingleLinkedList(Of GPXFile.OSMWayPoint) = GPXfile.GetWayPoints
                    Dim iconName As String
                    If wayPointList IsNot Nothing Then
                        Dim waypointBrush As Brush = New SolidBrush(waypointColor)
                        GPXfile.waypoint_mutex.WaitOne()
                        Try
                            Dim enumerator As IEnumerator(Of GPXFile.OSMWayPoint) = GPXfile.GetWayPoints.GetEnumerator
                            While enumerator.MoveNext
                                ' TODO: caching getXPPosition would increase performance
                                GetXYPosition(enumerator.Current.GetPosition, tileX, tileY, x_pos, y_pos)
                                If x_pos > 0 AndAlso y_pos > 0 AndAlso x_pos < nTileImage.Width AndAlso y_pos < nTileImage.Height Then
                                    If tileZ > 14 AndAlso paintPOIsEnabled Then
                                        iconName = "" ' future extension: store icon name in waypoint, resolved at first drawing
                                        OSMfile.drawIcon(g, iconName, x_pos, y_pos, enumerator.Current.Comment, enumerator.Current.Name)
                                    Else
                                        g.FillEllipse(waypointBrush, x_pos, y_pos, 3, 3)
                                    End If
                                End If
                            End While
                        Finally
                            GPXfile.waypoint_mutex.ReleaseMutex()
                        End Try
                    End If
                End If
            End If

            Try
                Me.Invoke(New myInvalidateDelegate(AddressOf Invalidate))
            Catch e As Exception
                ' ignore, might occour on shutdown
            End Try
            If WaitForMultipleObjects(2, multiHandler, 0, -1) = 0 Then
                ' stop handle signaled
                abort = True
            End If
        End While
        If g IsNot Nothing Then
            g.Dispose() 'is this required?
        End If
        LocalFree(multiHandler)
        CloseHandle(stopHandle)
        stopHandle = Nothing
        CloseHandle(repaintMapHandle)
        repaintMapHandle = Nothing
    End Sub
    ''' <summary>
    ''' Callback for the tile-downloader to inform the map to repaint the map after a tile was downloaded
    ''' </summary>
    ''' <param name="downloadedTile">The downloaded tile.</param>
    Private Sub UpdateTile(ByVal downloadedTile As OSMTileDownloader.TileQueueEntry) Handles downloader.DownloadedTile
        If downloadedTile.tileZ <> tileZ Then
            Exit Sub
        End If
        Dim found As Boolean = False
        tileCacheMutex.WaitOne()
        For i As Integer = 0 To tiles.GetLength(0) - 1
            For j As Integer = 0 To tiles.GetLength(1) - 1
                ' possible problem here if tile provider switched
                If tiles(i, j).CompareCoordinates(downloadedTile.tileX, downloadedTile.tileY, downloadedTile.tileZ) AndAlso tiles(i, j).IsAvailable = False Then
                    tiles(i, j).SetTileImage(RetrieveTile(downloadedTile.tileX, downloadedTile.tileY, downloadedTile.tileZ, Nothing))
                    found = True
                    Exit For
                End If
            Next
            If found = True Then
                Exit For
            End If
        Next
        tileCacheMutex.ReleaseMutex()

        If found = True Then
            forceUpdate = True
            UpdateMap()
        End If
    End Sub
    Public Delegate Sub myInvalidateDelegate()
    Private PowerOf2 As Integer() = {1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, _
                                         2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288, 1048576}

    ''' <summary>
    ''' Calculates the tile numbers for GPS position pos
    ''' </summary>
    ''' <param name="pos">The pos.</param>
    ''' <param name="zoom">The zoomlevel.</param>
    ''' <param name="x">The calculated tile x number.</param>
    ''' <param name="y">The calculated tile y number.</param>
    Private Sub CalcTileXY(ByVal pos As MAP_POSITION, ByVal zoom As Integer, ByRef x As Integer, ByRef y As Integer)
        ' http://dev.openstreetmap.org/~ojw/Tiles/tile.php/14/8452/5496.png

        Dim xf As Double = (pos.LonF + 180) / 360 * PowerOf2(zoom)
        Dim yf As Double = (1 - Math.Log(Math.Tan(pos.LatF * Math.PI / 180) + 1 / Math.Cos(pos.LatF * Math.PI / 180)) / Math.PI) / 2 * PowerOf2(zoom)
        x = CInt(Math.Floor(xf))
        y = CInt(Math.Floor(yf))
    End Sub
    ''' <summary>
    ''' Retrieves the tile z/x/y.
    ''' Checks if the tile is in the storage device and loads it. If not it enqueues it to download if downloading is enabled.
    ''' </summary>
    ''' <param name="x">The x tile number.</param>
    ''' <param name="y">The y tile number.</param>
    ''' <param name="z">The z tile number.</param>
    ''' <param name="downloader">The downloader to use.</param>
    ''' <returns>the loaded tile or null</returns>
    Private Function RetrieveTile(ByVal x As Integer, ByVal y As Integer, ByVal z As Integer, ByRef downloader As OSMTileDownloader) As Bitmap
        If currentTileProvider.FolderName IsNot Nothing AndAlso currentTileProvider.FolderName.Length > 0 Then
            Dim folder As String = currentTileProvider.FolderName & IO.Path.DirectorySeparatorChar & z & IO.Path.DirectorySeparatorChar & x

            ' support for bundled tiles
            If IO.File.Exists(folder + ".tiles") Then
                Dim inputStream As New IO.FileStream(folder + ".tiles", IO.FileMode.Open)
                Dim input As New IO.BinaryReader(inputStream)
                If New String(input.ReadChars(8)) = "OSMtTile" Then
                    input.ReadByte() ' zoom level
                    input.ReadInt32() ' x
                    While inputStream.Position < inputStream.Length
                        Dim readY As Integer = input.ReadInt32
                        input.ReadInt64() ' LastWriteTime
                        Dim length As Integer = input.ReadInt32
                        If readY = y Then
                            Dim byteArray(length - 1) As Byte
                            Dim thePicture As New IO.MemoryStream(length)
                            inputStream.Read(byteArray, 0, length)
                            thePicture.Write(byteArray, 0, length)
                            input.Close()
                            Return New Bitmap(thePicture)
                        ElseIf readY < y Then ' must be stored in ascending order
                            inputStream.Seek(length, IO.SeekOrigin.Current)
                        Else
                            Exit While
                        End If
                    End While
                End If
                input.Close()
            End If

            ' support for updatable tile-folder-format
            Dim fileName As String = folder & IO.Path.DirectorySeparatorChar & y & ".png"

            If Not IO.File.Exists(fileName) Then
                If downloadTilesEnabled AndAlso downloader IsNot Nothing AndAlso Not currentTileProvider.URL.Length = 0 Then
                    If Not IO.Directory.Exists(folder) Then
                        Try
                            IO.Directory.CreateDirectory(folder)
                        Catch ex As Exception
                            MsgBox("Error creating folder: " & folder, MsgBoxStyle.Critical)
                            ' do not enqeue tile for download if we cannot write
                            Return Nothing
                        End Try
                    End If
                    downloader.Download(New OSMTileDownloader.TileQueueEntry(currentTileProvider.URL & z & "/" & x & "/" & y & ".png", fileName, x, y, z))
                End If
            Else
                Try
                    Return New Bitmap(fileName)
                Catch ex As Exception
                End Try
            End If
        End If
        Return Nothing
    End Function
    ''' <summary>
    ''' Calculates the relative pixel coordinates of the position myPos from tile number tile_x x tile_y on the current zoomlevel
    ''' </summary>
    ''' <param name="myPos">My pos.</param>
    ''' <param name="x">The tile x number.</param>
    ''' <param name="y">The tile y number.</param>
    ''' <param name="xpos">The xpos in pixels "return" value.</param>
    ''' <param name="ypos">The ypos in pixels "return" value.</param>
    Private Sub GetXYPosition(ByVal myPos As MAP_POSITION, ByVal x As Integer, ByVal y As Integer, ByRef xpos As Integer, ByRef ypos As Integer)
        Dim xf As Double = (myPos.LonF + 180.0) / 360.0 * PowerOf2(tileZ)
        Dim yf As Double = (1 - Math.Log(Math.Tan(myPos.LatF * Math.PI / 180.0) + 1 / Math.Cos(myPos.LatF * Math.PI / 180.0)) / Math.PI) / 2.0 * PowerOf2(tileZ)

        ' get the difference
        xf = xf - x
        yf = yf - y

        ' translate to tile position
        xpos = CInt(xf * TILE_WIDTH)
        ypos = CInt(yf * TILE_HEIGHT)
    End Sub
    ''' <summary>
    ''' Draws the own position with an arrow representing the direction/heading on the map.
    ''' </summary>
    ''' <param name="e">The Graphic to paint on.</param>
    ''' <param name="x">The x position in pixel.</param>
    ''' <param name="y">The y position in pixel.</param>
    Private Sub DrawPosition(ByVal e As Graphics, ByVal x As Integer, ByVal y As Integer)
        Const DegPI As Double = Math.PI / 180.0
        Dim BearingRad As Double = heading * DegPI
        Dim sinRad As Double = Math.Sin(BearingRad)
        Dim cosRad As Double = Math.Cos(BearingRad)
        Dim points(3) As Point

        Dim hsinRad As Integer = CInt(arrowHeight * sinRad)
        Dim wsinRad As Integer = CInt(arrowWidth * sinRad)
        Dim hcosRad As Integer = CInt(arrowHeight * cosRad)
        Dim wcosRad As Integer = CInt(arrowWidth * cosRad)

        points(0) = New Point(x + wsinRad, y - hcosRad)
        points(1) = New Point(x - wcosRad - hsinRad, y - wsinRad + hcosRad)
        points(2) = New Point(x + wsinRad - hsinRad, y)
        points(3) = New Point(x + wcosRad - hsinRad, y + wsinRad + hcosRad)

        If manualMode = False Then
            e.DrawPolygon(New Pen(Color.DarkRed, 2), points)
            e.FillPolygon(New SolidBrush(Color.Red), points)
        Else
            e.DrawPolygon(New Pen(Color.DarkBlue, 2), points)
            e.FillPolygon(New SolidBrush(Color.LightBlue), points)
        End If
    End Sub
    Public Sub PaintWayPoint()
        'Graphics.FromImage(nTileImage).FillEllipse(New SolidBrush(Color.Blue), x_pos, y_pos, 10, 10)
        paintWayPointsEnabled = True
        UpdateMap()
    End Sub
    ''' <summary>
    ''' Switches the tile provider to the specified one and updates the map.
    ''' </summary>
    ''' <param name="tileProvider">The tile provider.</param>
    ''' <param name="updateDisplay">if set to <c>true</c> [update display].</param>
    Public Sub UpdateTileProvider(ByVal tileProvider As TileProviders.TileProvider, Optional ByVal updateDisplay As Boolean = False)
        Dim updated As Boolean = False
        If Not tileProvider.URL = currentTileProvider.FolderName Then
            updated = True
        ElseIf Not tileProvider.FolderName = currentTileProvider.FolderName Then
            updated = True
        End If
        currentTileProvider = tileProvider
        If tileZ > currentTileProvider.MaxZoom Then
            tileZ = currentTileProvider.MaxZoom
        ElseIf tileZ < currentTileProvider.MinZoom Then
            tileZ = currentTileProvider.MinZoom
        End If
        RaiseEvent Zoomed()
        If updateDisplay AndAlso updated Then
            tileCacheMutex.WaitOne()
            ' force update of the map and kill the tile cache
            For i As Integer = 0 To tiles.GetLength(0) - 1
                For j As Integer = 0 To tiles.GetLength(1) - 1
                    tiles(i, j) = New OSMTile()
                Next j
            Next i
            tileCacheMutex.ReleaseMutex()
            forceUpdate = True
            UpdateMap()
        End If
    End Sub
    ''' <summary>
    ''' Updates the map.
    ''' </summary>
    Private Sub UpdateMap()
        If repaintSlippyMapThread Is Nothing Then
            repaintSlippyMapThread = New Threading.Thread(New Threading.ThreadStart(AddressOf PrepareSlippyMapThread))
            stopHandle = CreateEvent(IntPtr.Zero, 0, 0, Nothing)
            repaintMapHandle = CreateEvent(IntPtr.Zero, 0, 0, Nothing)
            repaintSlippyMapThread.Priority = Threading.ThreadPriority.BelowNormal
            repaintSlippyMapThread.Start()
        Else
            SetEvent(repaintMapHandle)
        End If
    End Sub
    ''' <summary>
    ''' Scrolls the specified direction.
    ''' </summary>
    ''' <param name="direction">The direction.</param>
    Public Sub Scroll(ByVal direction As ScrollDirection)
        Dim diff As Double = 0.0005
        If tileZ <= 2 Then
            diff = 6
        ElseIf tileZ < 4 Then
            diff = 3
        ElseIf tileZ < 6 Then
            diff = 0.5
        ElseIf tileZ < 8 Then
            diff = 0.1
        ElseIf tileZ < 10 Then
            diff = 0.05
        ElseIf tileZ < 12 Then
            diff = 0.01
        ElseIf tileZ < 15 Then
            diff = 0.001
        End If
        EnterManualMode()
        Select Case direction
            Case ScrollDirection.DOWN
                mapPosition.ChangeDelta(-diff, 0)
            Case ScrollDirection.UP
                mapPosition.ChangeDelta(diff, 0)
            Case ScrollDirection.LEFT
                mapPosition.ChangeDelta(0, -diff)
            Case ScrollDirection.RIGHT
                mapPosition.ChangeDelta(0, diff)
        End Select
        UpdateMap()
    End Sub
    ''' <summary>
    ''' Zooms in the map.
    ''' </summary>
    Public Sub ZoomIn()
        If IsZoomInAble() Then
            tileZ += 1
            downloader.ClearQueue()
            UpdateMap()
            RaiseEvent Zoomed()
        End If
    End Sub
    ''' <summary>
    ''' Zooms out the map.
    ''' </summary>
    Public Sub ZoomOut()
        If IsZoomOutAble() Then
            tileZ -= 1
            downloader.ClearQueue()
            UpdateMap()
            RaiseEvent Zoomed()
        End If
    End Sub
    ''' <summary>
    ''' Enters the manual (scrolling) mode.
    ''' </summary>
    Public Sub EnterManualMode()
        If manualMode = False Then
            mapPosition = gpsPosition
            manualMode = True
        End If
    End Sub
    ''' <summary>
    ''' Exits the manual (scrolling) mode.
    ''' </summary>
    Public Sub ExitManualMode()
        If manualMode = True Then
            manualMode = False
            UpdateMap()
        End If
    End Sub
    ''' <summary>
    ''' Determines whether zoom in is possible with the current map/tile-provider.
    ''' </summary>
    ''' <returns>
    ''' <c>true</c> if zoom in is possible; otherwise, <c>false</c>.
    ''' </returns>
    Public Function IsZoomInAble() As Boolean
        Return (tileZ < currentTileProvider.MaxZoom)
    End Function
    ''' <summary>
    ''' Determines whether zoom out is possible with the current map/tile-provider.
    ''' </summary>
    ''' <returns>
    ''' <c>true</c> if zoom out is possible; otherwise, <c>false</c>.
    ''' </returns>
    Public Function IsZoomOutAble() As Boolean
        Return (tileZ > currentTileProvider.MinZoom)
    End Function
    ''' <summary>
    ''' Shuts down this instance and stops the background thread for updating the nTimeImage
    ''' </summary>
    Sub Shutdown()
        If stopHandle <> IntPtr.Zero Then
            SetEvent(stopHandle)
        End If
        ' disable downloads in order to clear queue and make osmtracker quit
        DownloadTiles = False
    End Sub
    Private Sub osmTileMap_MouseDown(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles MyBase.MouseDown
        dragging = True
        startX = e.X
        startY = e.Y

        If manualMode = True Then
            ' if manual start in current map position
            startPos = Projection.LatLon2Mercator(New PointD(mapPosition.LonF, mapPosition.LatF))
        Else
            ' if automatic start in current GPS position
            startPos = Projection.LatLon2Mercator(New PointD(gpsPosition.LonF, gpsPosition.LatF))
        End If
    End Sub
    Private Sub osmTileMap_MouseUp(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles MyBase.MouseUp
        dragging = False
    End Sub
    Private Sub osmTileMap_MouseMove(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles MyBase.MouseMove
        If dragging Then
            Dim deltaX As Integer = startX - e.X
            Dim deltaY As Integer = e.Y - startY
            Dim diff As Double
            If tileZ >= MIN_ZOOM AndAlso tileZ < MAX_ZOOM Then
                diff = MercUnitPerPixel(tileZ)
            End If

            If deltaX <> 0 OrElse deltaY <> 0 Then
                EnterManualMode()
                startPos.Change(deltaX * diff, deltaY * diff)
                Dim p As PointD = Projection.Mercator2LatLon(startPos)
                mapPosition.Change(p.y, p.x)

                startX = e.X
                startY = e.Y

                UpdateMap()
            End If
        End If
    End Sub
    Private Shared MercUnitPerPixel() As Single = {156543.339, 78271.51695, 39135.75848, 19567.87924, 9783.939619, _
                                               4891.96981, 2445.984905, 1222.992452, 611.4962262, _
                                               305.7481131, 152.8740565, 76.43702827, 38.21851414, _
                                               19.10925707, 9.554628534, 4.777314267, 2.388657134, _
                                               1.194328567, 0.597164283, 0.298582142}

    Public Sub New()
        InitializeComponent()
        Try
            blankTile = New Bitmap(GetAppPath() & IO.Path.DirectorySeparatorChar & "img" & IO.Path.DirectorySeparatorChar & "tilesync.png")
        Catch e As Exception
            blankTile = New Bitmap(TILE_WIDTH, TILE_HEIGHT, Imaging.PixelFormat.Format16bppRgb555)
            Dim g As Graphics = Graphics.FromImage(blankTile)
            g.FillRectangle(New SolidBrush(Color.LightGray), 0, 0, TILE_WIDTH, TILE_HEIGHT)
            g.DrawLines(New Pen(Color.Black, 3), New Point() {New Point(100, 95), New Point(150, 95), New Point(100, 155), New Point(150, 155), New Point(100, 95)})
            g.Dispose()
        End Try
    End Sub
End Class
