' Microsoft .NET Compact Framework-based PInvoke Library Sample
' Licensed under the Shared Source License for Microsoft .NET Compact Framework-based PInvoke Library Sample
'
' based on http://msdn.microsoft.com/en-us/library/aa446550.aspx
'
' Geoff Schwab
' Excell Data Corporation
'
' Contributors:
'    Jonathan Wells
'    Microsoft Corporation
'
' Adapted for OSMtracker by:
'     Sven Strickroth
'
' Wave Samples:
'    Seth Demse
'    Microsoft Corporation
'
'    Jonathan Wells
'    Microsoft Corporation

Imports System
Imports System.Runtime.InteropServices
Imports System.IO
Imports Microsoft.WindowsCE.Forms
Imports System.Text
Imports System.Threading
Imports System.Diagnostics

'/ <summary>
'/ Encapsulates Waveform Audio Interface recording functions and provides a simple
'/ interface for recording audio.
'/ </summary>
Public Class WaveIn

    ' <summary>
    ' Supplies an inteface for creating, recording to, and saving a .wav file.
    ' </summary>
    Protected Class WaveFile
        Implements IDisposable

        Private desiredBitsPerSample As Integer
        Private desiredSamplesPerSecond As Integer

        ' <summary>
        ' Hardware interface instance for this wave file.
        ' </summary>
        Protected m_hwi As IntPtr = IntPtr.Zero

        ' <summary>
        ' Instance of the WAVEFORMATEX header for this file.
        ' </summary>
        Protected m_wfmt As Wave.WAVEFORMATEX = Nothing

        ' <summary>
        ' Buffers used to store recording information.
        ' </summary>
        Protected m_whdr() As Wave.WAVEHDR = Nothing

        ' <summary>
        ' Specifies whether the file is inited or not.
        ' </summary>
        Protected m_inited As Boolean = False

        ' <summary>
        ' Current block, or buffer in m_whdr, being written.
        ' </summary>
        Protected m_curBlock As Integer

        ' <summary>
        ' Maximum number of blocks to be written.
        ' </summary>
        Protected m_numBlocks As Integer

        ' <summary>
        ' Size of each record buffer.
        ' </summary>
        Protected m_bufferSize As Integer

        ' <summary>
        ' Maximum size of all record buffers.
        ' </summary>
        Protected m_maxDataLength As Integer

        ' <summary>
        ' Specifies whether or not recording is done.
        ' </summary>
        Public ReadOnly Property Done() As Boolean
            Get
                Return Not m_recording
            End Get
        End Property
        Protected m_recording As Boolean = False

        ' <summary>
        ' Preload the buffers and prepare them for recording.
        ' </summary>
        ' <param name="curDevice">Device to use for recording</param>
        ' <param name="hwnd">Handle to a message window that will receive
        ' audio messages.</param>
        ' <param name="maxRecordLength_ms">Maximum length of recording</param>
        ' <param name="bufferSize">Size of buffers to use for recording.  New
        ' buffers are added when needed until the maximum length is reached
        ' or the recording is stopped.</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function Preload(ByVal curDevice As Integer, ByVal hwnd As IntPtr, ByVal maxRecordLength_ms As Integer, ByVal bufferSize As Integer) As Wave.MMSYSERR

            ' Do not allow recording to be interrupted
            If m_recording Then
                Return Wave.MMSYSERR.GENERROR
            End If

            ' If this file is already initialized then start over
            If m_inited Then
                StopRecord()
                FreeWaveBuffers()
            End If

            ' Create an instance of WAVEINCAPS to check if our desired
            ' format is supported
            Dim caps As WAVEINCAPS = New WAVEINCAPS
            waveInGetDevCaps(0, caps.Data, caps.Size)
            If (caps.dwFormats And Wave.WAVE_FORMAT_2S16) = 0 Then
                Return Wave.MMSYSERR.NOTSUPPORTED
            End If

            ' Initialize a WAVEFORMATEX structure specifying the desired
            ' format
            m_wfmt = New Wave.WAVEFORMATEX
            m_wfmt.wFormatTag = Wave.WAVEHDR.WAVE_FORMAT_PCM
            m_wfmt.wBitsPerSample = desiredBitsPerSample
            m_wfmt.nChannels = 1
            m_wfmt.nSamplesPerSec = desiredSamplesPerSecond
            m_wfmt.nAvgBytesPerSec = CInt((m_wfmt.nSamplesPerSec * m_wfmt.nChannels * m_wfmt.wBitsPerSample) >> 3)
            m_wfmt.nBlockAlign = CInt((m_wfmt.wBitsPerSample * m_wfmt.nChannels) >> 3)

            ' Attempt to open the specified device with the desired wave format
            Dim result As Wave.MMSYSERR = waveInOpen(m_hwi, curDevice, m_wfmt, hwnd, 0, Wave.CALLBACK_WINDOW)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            If bufferSize = 0 Then
                Return Wave.MMSYSERR.GENERROR
            End If

            m_bufferSize = bufferSize

            ' Force the buffers to align to nBlockAlign
            If (m_bufferSize Mod m_wfmt.nBlockAlign) <> 0 Then
                m_bufferSize += m_wfmt.nBlockAlign - (m_bufferSize Mod m_wfmt.nBlockAlign)
            End If

            ' Determine the number of buffers needed to record the maximum length
            m_maxDataLength = CInt((m_wfmt.nAvgBytesPerSec * maxRecordLength_ms) / 1000)
            m_numBlocks = CInt(Fix(m_maxDataLength / m_bufferSize))
            If (m_numBlocks * m_bufferSize) < m_maxDataLength Then
                m_numBlocks += 1
            End If

            ' Allocate the list of buffers
            m_whdr = New Wave.WAVEHDR(m_numBlocks) {}


            ' Allocate and initialize two buffers to start with
            m_whdr(0) = New Wave.WAVEHDR
            m_whdr(1) = New Wave.WAVEHDR

            result = InitBuffer(0)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            result = InitBuffer(1)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            m_curBlock = 0
            m_inited = True

            Return Wave.MMSYSERR.NOERROR

        End Function

        ' <summary>
        ' Called when the Windows message specifying that a block has finished
        ' recording is encountered.  It is critical that the application stay
        ' one buffer ahead because at the point this function is called, the system
        ' has already started writing to the next buffer.
        ' </summary>
        Public Sub BlockDone()

            m_curBlock += 1

            ' If the next block is not the padding buffer at the end of the
            ' recording then initialize another buffer, otherwise stop recording
            If m_curBlock < m_numBlocks Then
                InitBuffer(m_curBlock + 1)
            ElseIf m_curBlock = m_numBlocks Then
                StopRecord()
            End If

        End Sub

        ' <summary>
        ' Initialize a buffer for recording by allocating a data buffer,
        ' preparing the header, and adding it to the read queue.
        ' </summary>
        ' <param name="bufIndex">Index of buffer to allocate</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function InitBuffer(ByVal bufIndex As Integer) As Wave.MMSYSERR

            ' Determine the size of the buffer to create
            Dim writeLength As Integer = m_bufferSize

            If bufIndex < m_numBlocks Then
                Dim remainingDataLength As Integer = m_maxDataLength - (bufIndex * m_bufferSize)
                If m_bufferSize > remainingDataLength Then
                    writeLength = remainingDataLength
                End If
            End If

            ' If the header is not already instanced then instance it
            If m_whdr(bufIndex) Is Nothing Then
                m_whdr(bufIndex) = New Wave.WAVEHDR
            End If

            ' Allocate memory if not already allocated
            Dim result As Wave.MMSYSERR = m_whdr(bufIndex).Init(writeLength, False)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            ' Prepare the header
            result = waveInPrepareHeader(m_hwi, m_whdr(bufIndex), Marshal.SizeOf(m_whdr(bufIndex)))
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            ' Put the buffer in the queue
            Return waveInAddBuffer(m_hwi, m_whdr(bufIndex), Marshal.SizeOf(m_whdr(bufIndex)))

        End Function

        ' <summary>
        ' Start recording.
        ' </summary>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function Start() As Wave.MMSYSERR

            If Not m_inited OrElse m_recording Then
                Return Wave.MMSYSERR.GENERROR
            End If

            Dim result As Wave.MMSYSERR = waveInStart(m_hwi)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            m_recording = True

            Return Wave.MMSYSERR.NOERROR

        End Function

        ' <summary>
        ' Free the buffers allocated for recording.  This is not done when
        ' recording stops because the user needs a chance to save or reject
        ' the recording before trying again.
        ' </summary>
        Private Sub FreeWaveBuffers()

            m_inited = False

            If Not (m_whdr Is Nothing) Then

                Dim i As Integer
                For i = 0 To m_whdr.Length - 1
                    If Not (m_whdr(i) Is Nothing) Then
                        waveInUnprepareHeader(m_hwi, m_whdr(i), Marshal.SizeOf(m_whdr(i)))

                        m_whdr(i).Dispose()
                        m_whdr(i) = Nothing
                    End If
                Next

                m_whdr = Nothing

            End If

            waveInClose(m_hwi)

            m_hwi = IntPtr.Zero

        End Sub

        ' <summary>
        ' Write individual characters to the BinaryWriter.  This is a helper
        ' function used by Save.  Writing strings does not work because they
        ' write a length before the characters.
        ' </summary>
        ' <param name="wrtr">Destination of write</param>
        ' <param name="text">Characters to be written</param>
        Private Sub WriteChars(ByVal wrtr As BinaryWriter, ByVal txt As String)

            Dim i As Integer
            For i = 0 To txt.Length - 1
                Dim c As Char = txt.Chars(i)
                wrtr.Write(c)
            Next

        End Sub

        ' <summary>
        ' Save the current record buffers to the specified file.
        ' </summary>
        ' <param name="fileName">Name of file to be saved</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function Save(ByVal fileName As String) As Wave.MMSYSERR

            If Not m_inited Then
                Return Wave.MMSYSERR.GENERROR
            End If

            If m_recording Then
                StopRecord()
            End If

            Dim strm As FileStream = Nothing
            Dim wrtr As BinaryWriter = Nothing

            Try

                If File.Exists(fileName) Then

                    Dim fi As FileInfo = New FileInfo(fileName)
                    If (fi.Attributes And FileAttributes.ReadOnly) <> 0 Then
                        fi.Attributes = fi.Attributes Xor FileAttributes.ReadOnly
                    End If

                    strm = New FileStream(fileName, FileMode.Truncate)
                Else
                    strm = New FileStream(fileName, FileMode.Create)
                End If

                If strm Is Nothing Then
                    Return Wave.MMSYSERR.GENERROR
                End If

                wrtr = New BinaryWriter(strm)
                If wrtr Is Nothing Then
                    Return Wave.MMSYSERR.GENERROR
                End If

                ' Determine the size of the data, as the total number of bytes
                ' recorded by each buffer
                Dim totalSize As Integer = 0
                Dim i As Integer
                For i = 0 To m_numBlocks - 1
                    If Not (m_whdr(i) Is Nothing) Then
                        totalSize += m_whdr(i).dwBytesRecorded
                    End If
                Next


                Dim chunkSize As Integer = 36 + totalSize

                ' Write out the header information
                WriteChars(wrtr, "RIFF")
                wrtr.Write(chunkSize)
                WriteChars(wrtr, "WAVEfmt ")
                wrtr.Write(Convert.ToInt32(16))
                m_wfmt.Write(wrtr)
                WriteChars(wrtr, "data")
                wrtr.Write(totalSize)

                ' Write the data recorded to each buffer
                For i = 0 To m_numBlocks - 1
                    If Not (m_whdr(i) Is Nothing) Then
                        Dim result As Wave.MMSYSERR = m_whdr(i).Write(wrtr)
                        If result <> Wave.MMSYSERR.NOERROR Then
                            Return result
                        End If
                    End If
                Next

                Return Wave.MMSYSERR.NOERROR

            Finally
                FreeWaveBuffers()

                If Not (strm Is Nothing) Then
                    strm.Close()
                End If

                If Not (wrtr Is Nothing) Then
                    wrtr.Close()
                End If

            End Try

        End Function

        ' <summary>
        ' Stop recording.  After stopping, the buffers are kept until recording
        ' is restarted or the buffers are saved.
        ' </summary>
        Public Sub StopRecord()

            waveInReset(m_hwi)
            m_recording = False

        End Sub

        ' <summary>
        ' Clean up any allocated resources.
        ' </summary>
        Public Sub Dispose() Implements IDisposable.Dispose

            StopRecord()
            FreeWaveBuffers()

        End Sub
        Public Sub New(ByVal bitsPerSecond As Integer, ByVal samplesPerSecond As Integer)
            desiredBitsPerSample = bitsPerSecond
            desiredSamplesPerSecond = samplesPerSecond
        End Sub
    End Class

    ' <summary>
    ' Defines the MessageWindow used to receive messages from the audio
    ' system.
    ' </summary>
    Protected Class SoundMessageWindow
        Inherits MessageWindow

        Public Const MM_WIM_OPEN As Integer = &H3BE
        Public Const MM_WIM_CLOSE As Integer = &H3BF
        Public Const MM_WIM_DATA As Integer = &H3C0

        ' Instance of a recording interface
        Protected m_wi As WaveIn = Nothing

        Public Sub New(ByVal wi As WaveIn)
            m_wi = wi
        End Sub

        Protected Overrides Sub WndProc(ByRef msg As Message)

            Select Case (msg.Msg)
                ' When this message is encountered, a block is
                ' done recording, so notify the WaveIn instance.
            Case MM_WIM_DATA
                    If Not (m_wi Is Nothing) Then
                        m_wi.BlockDone()
                    End If
            End Select

            MyBase.WndProc(msg)

        End Sub

    End Class


    ' <summary>
    ' Maintain an instance of a MessageWindow that handles audio messages.
    ' </summary>
    Protected m_msgWindow As SoundMessageWindow = Nothing

    ' <summary>
    ' An instance of WaveFile used as the destination for recording audio.
    ' </summary>
    Protected m_file As WaveFile = Nothing

    ' <summary>
    ' Create an instance of WaveIn.
    ' </summary>
    Public Sub New(ByVal bitsPerSecond As Integer, ByVal samplesPerSecond As Integer)
        m_msgWindow = New SoundMessageWindow(Me)
        m_file = New WaveFile(bitsPerSecond, samplesPerSecond)
    End Sub

    ' <summary>
    '  Determine the number of available recording devices.
    ' </summary>
    ' <returns>Number of input devices</returns>
    Public Function NumDevices() As Integer
        Return waveInGetNumDevs()
    End Function

    ' <summary>
    ' Get the name of the specified recording device.
    ' </summary>
    ' <param name="deviceId">ID of the device</param>
    ' <param name="prodName">Destination string assigned the name</param>
    ' <returns>MMSYSERR.NOERROR if successful</returns>
    Public Function GetDeviceName(ByVal deviceId As Integer, ByRef prodName As String) As Wave.MMSYSERR
        Dim caps As WAVEINCAPS = New WAVEINCAPS
        Dim result As Wave.MMSYSERR = waveInGetDevCaps(deviceId, caps.Data, caps.Size)

        If result <> Wave.MMSYSERR.NOERROR Then
            Return result
        End If

        prodName = caps.szPname

        Return Wave.MMSYSERR.NOERROR

    End Function

    ' <summary>
    ' A block has finished so notify the WaveFile.  In a more complicated example,
    ' this class might maintain an array of WaveFile instances.  In such a case, the
    ' wParam of the message could be passed from the MM_WIM_DATA message.  This
    ' value represents the m_hwi member of the file that caused the message.
    ' The code might look something like:
    '    Dim f As WaveFile
    '    For Each f In  m_files
    '        If f.m_hwi.ToInt32() = wParam.ToInt32() Then
    '            f.BlockDone();
    '            Exit ForEach
    '        End If
    '    Next f
    ' </summary>
    Public Sub BlockDone()
        m_file.BlockDone()
    End Sub

    ' <summary>
    ' Preload the buffers of the record file.
    ' </summary>
    ' <param name="maxRecordLength_ms">Maximum record length in milliseconds</param>
    ' <param name="bufferSize">Size of individual buffers, in bytes</param>
    ' <returns>MMSYSERR.NOERROR if successful</returns>
    Public Function Preload(ByVal maxRecordLength_ms As Integer, ByVal bufferSize As Integer) As Wave.MMSYSERR

        If Not (m_file Is Nothing) Then
            Return m_file.Preload(0, m_msgWindow.Hwnd, maxRecordLength_ms, bufferSize)
        End If

        Return Wave.MMSYSERR.NOERROR

    End Function

    ' <summary>
    ' Stop recording.
    ' </summary>
    Public Sub StopRecord()

        If Not (m_file Is Nothing) Then
            m_file.StopRecord()
        End If

    End Sub

    ' <summary>
    ' Start recording.
    ' </summary>
    ' <returns>MMSYSERR.NOERROR if successful</returns>
    Public Function Start() As Wave.MMSYSERR

        If Not (m_file Is Nothing) Then
            Return m_file.Start()
        End If

        Return Wave.MMSYSERR.NOERROR

    End Function

    ' <summary>
    ' Checks if the recording time is expired.
    ' </summary>
    ' <returns>true if not currently recording</returns>
    Public Function Done() As Boolean
        Return m_file.Done
    End Function

    ' <summary>
    ' Save the WaveFile buffers to the specified file.
    ' </summary>
    ' <param name="fileName">Name of destination file</param>
    ' <returns>MMSYSERR.NOERROR if successful</returns>
    Public Function Save(ByVal fileName As String) As Wave.MMSYSERR

        If Not (m_file Is Nothing) Then
            Return m_file.Save(fileName)
        End If

        Return Wave.MMSYSERR.NOERROR

    End Function

    ' <summary>
    ' Clean up any resources allocated by the class.
    ' </summary>
    Public Sub Dispose()

        m_msgWindow.Dispose()

        If Not (m_file Is Nothing) Then
            m_file.Dispose()
        End If

    End Sub

    ' <summary>
    ' This function retrieves the number of waveform input devices present in the system.
    ' </summary>
    ' <returns>The number of devices indicates success. Zero indicates that no devices are present or that an error occurred.</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveInGetNumDevs() As Integer
    End Function

    ' <summary>
    ' This function opens a specified waveform input device for recording.
    ' </summary>
    ' <param name="phwi">Address filled with a handle identifying the open
    ' waveform-audio input device. Use this handle to identify the device when
    ' calling other waveform-audio input functions. This parameter can be NULL
    ' if WAVE_FORMAT_QUERY is specified for fdwOpen.</param>
    ' <param name="uDeviceID">Identifier of the waveform-audio input device to open.
    ' It can be either a device identifier or a Handle to an open waveform-audio
    ' input device. Can also be WAVE_MAPPER.</param>
    ' <param name="pwfx">Pointer to a WAVEFORMATEX structure that identifies the
    ' desired format for recording waveform-audio data. You can free this structure
    ' immediately after waveInOpen returns.</param>
    ' <param name="dwCallback">Specifies the address of a fixed callback function,
    ' an event handle, a handle to a window, or the identifier of a thread to be
    ' called during waveform-audio recording to process messages related to the
    ' progress of recording. If no callback function is required, this value can be
    ' zero.</param>
    ' <param name="dwInstance">Specifies user-instance data passed to the callback
    ' mechanism. This parameter is not used with the window callback mechanism.</param>
    ' <param name="fdwOpen">Flags for opening the device.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Private Shared Function waveInOpen(ByRef phwi As IntPtr, ByVal uDeviceID As Integer, ByVal pwfx As Wave.WAVEFORMATEX, ByVal dwCallback As IntPtr, ByVal dwInstance As Integer, ByVal fdwOpen As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function prepares a buffer for waveform input.
    ' </summary>
    ' <param name="hwi">Handle to the waveform-audio input device.</param>
    ' <param name="pwh">Pointer to a WAVEHDR structure that identifies the buffer
    ' to be prepared. The buffer's base address must be aligned with the respect
    ' to the sample size.</param>
    ' <param name="cbwh">Size, in bytes, of the WAVEHDR structure.</param>
    ' <returns></returns>
    <DllImport("coredll.dll")> _
    Private Shared Function waveInPrepareHeader(ByVal hwi As IntPtr, ByVal pwh As Wave.WAVEHDR, ByVal cbwh As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function cleans up the preparation performed by waveInPrepareHeader.
    ' The function must be called after the device driver fills a data buffer
    ' and returns it to the application. You must call this function before
    ' freeing the data buffer.
    ' </summary>
    ' <param name="hwi">Handle to the waveform-audio input device.</param>
    ' <param name="pwh">Pointer to a WAVEHDR structure identifying the buffer to
    ' be cleaned up.</param>
    ' <param name="cbwh">Size, in bytes, of the WAVEHDR structure.</param>
    ' <returns></returns>
    <DllImport("coredll.dll")> _
    Private Shared Function waveInUnprepareHeader(ByVal hwi As IntPtr, ByVal pwh As Wave.WAVEHDR, ByVal cbwh As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function closes the specified waveform-audio input device.
    ' </summary>
    ' <param name="hwi">Handle to the waveform-audio input device. If the
    ' function succeeds, the handle is no longer valid after this call.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveInClose(ByVal hwi As IntPtr) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function stops input on a specified waveform input device and resets
    ' the current position to 0. All pending buffers are marked as done and
    ' returned to the application.
    ' </summary>
    ' <param name="hwi">Handle to the waveform-audio input device.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveInReset(ByVal hwi As IntPtr) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function starts input on the specified waveform input device.
    ' </summary>
    ' <param name="hwi">Handle to the waveform-audio input device.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveInStart(ByVal hwi As IntPtr) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function stops waveform input.
    ' </summary>
    ' <param name="hwi">Handle to the waveform-audio input device.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveInStop(ByVal hwi As IntPtr) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function sends an input buffer to the specified waveform-audio input
    ' device. When the buffer is filled, the application is notified.
    ' </summary>
    ' <param name="hwi">Handle to the waveform-audio input device.</param>
    ' <param name="pwh">Pointer to a WAVEHDR structure that identifies
    ' the buffer.</param>
    ' <param name="cbwh">Size, in bytes, of the WAVEHDR structure.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Private Shared Function waveInAddBuffer(ByVal hwi As IntPtr, ByVal pwh As Wave.WAVEHDR, ByVal cbwh As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This structure describes the capabilities of a waveform-audio input device.
    '		typedef struct 
    '		{
    '			WORD wMid; 
    '			WORD wPid; 
    '			MMVERSION vDriverVersion; 
    '			TCHAR szPname[MAXPNAMELEN]; 
    '			DWORD dwFormats; 
    '			WORD wChannels; 
    '			WORD wReserved1;} 
    '		WAVEINCAPS;
    '	This structure has an embedded TCHAR array so the managed implementation is
    '	a byte array with accessors.
    ' </summary>
    Protected Class WAVEINCAPS

        Const WAVEINCAPS_SIZE As Integer = 80

        Private m_data() As Byte = Nothing

        Public ReadOnly Property Size() As Integer
            Get
                Return WAVEINCAPS_SIZE
            End Get
        End Property

        ' <summary>
        ' Manufacturer identifier for the device driver for the waveform-audio
        ' input device.
        ' </summary>
        Public ReadOnly Property wMid() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt16(m_data, 0))
            End Get
        End Property
        ' <summary>
        ' Product identifier for the waveform-audio input device. Product
        ' identifiers are defined in Manufacturer and Product Identifiers.
        ' </summary>
        Public ReadOnly Property wPid() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt16(m_data, 2))
            End Get
        End Property
        ' <summary>
        ' Version number of the device driver for the waveform-audio input device.
        ' The high-order byte is the major version number, and the low-order byte
        ' is the minor version number.
        ' </summary>
        Public ReadOnly Property vDriverVersion() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt32(m_data, 4))
            End Get
        End Property
        ' <summary>
        ' Specifies the standard formats that are supported. It is one or
        ' a combination of the following flags.
        ' </summary>
        Public ReadOnly Property dwFormats() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt32(m_data, 72))
            End Get
        End Property
        ' <summary>
        ' Number that specifies whether the device supports mono (1) or stereo (2)
        ' input.
        ' </summary>
        Public ReadOnly Property wChannels() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt16(m_data, 76))
            End Get
        End Property
        ' <summary>
        ' Padding.
        ' </summary>
        Public ReadOnly Property wReserved1() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt16(m_data, 78))
            End Get
        End Property

        ' <summary>
        ' Null-terminated string that contains the product name.
        ' </summary>
        Public ReadOnly Property szPname() As String
            Get
                Dim bytes(31) As Char
                Dim i As Integer
                For i = 0 To 31
                    bytes(i) = Convert.ToChar(BitConverter.ToUInt16(m_data, i * 2 + 8))
                Next

                Return New String(bytes)

            End Get
        End Property

        Public Sub New()
            m_data = New Byte(WAVEINCAPS_SIZE - 1) {}
        End Sub

        Public ReadOnly Property Data() As Byte()
            Get
                Return m_data
            End Get
        End Property

    End Class

    ' <summary>
    ' This function retrieves the capabilities of a specified waveform-audio
    ' input device.
    ' </summary>
    ' <param name="uDeviceID">Identifier of the waveform-audio output device. It
    ' can be either a device identifier or a Handle to an open waveform-audio
    ' input device.</param>
    ' <param name="pwic">Pointer to a WAVEINCAPS structure to be filled with
    ' information about the capabilities of the device.</param>
    ' <param name="cbwic">Size, in bytes, of the WAVEINCAPS structure.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveInGetDevCaps(ByVal uDeviceID As Integer, ByVal pwic() As Byte, ByVal cbwic As Integer) As Wave.MMSYSERR
    End Function


End Class 'WaveIn